// ==========================================================================
//  Copyright  2006 Boguslaw Cyganek.  All rights reserved.
// ==========================================================================
//
//
// ==========================================================================
//  History:
// ==========================================================================
//
//  
//
// ==========================================================================


#ifndef MEMORY_MAPPED_FILE_ALLOCATOR_H
#define MEMORY_MAPPED_FILE_ALLOCATOR_H




#include "Windows.h"


#include <String>
using namespace std;



///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////
// Template declaration.

template < class AType > class TMemoryMappedFileAllocator
{
	public:

		virtual ~TMemoryMappedFileAllocator() {}			// class virtual destructor

	public:
		AType * Allocate( LPSTR fileName, DWORD howToCreate = OPEN_EXISTING, DWORD shareMode = FILE_SHARE_READ | FILE_SHARE_WRITE /*| FILE_SHARE_DELETEadded by BC on July 30th'98*/, DWORD initialSize = 0 );

		void	Free( AType * mappedPtr );

		DWORD	fAllocationSize;
};




///////////////////////////////////////////////////////////////////////////////////////
// Template implementation.

// Use it to create and map the picture file into your address space.
// It uses the CreateFile with OPEN_ALWAYS parameter.
// INPUT:	fileName - a pointer to the string containing file
//			fileSize - contains size of a file to be created
//			howToCreate -	specifies the type of file creation(CREATE_ALWAYS is default value), 
//							possible values are:
//
//							
//							CREATE_NEW	Creates a new file. 
//										The function fails if the specified file already exists.
//							CREATE_ALWAYS	Creates a new file. 
//										The function overwrites the file if it exists.
//							OPEN_EXISTING	Opens the file. 
//									The function fails if the file does not exist.
//							OPEN_ALWAYS	Opens the file, if it exists. 
//										If the file does not exist, the function creates the file 
//										as if fdwCreate were CREATE_NEW.
//							TRUNCATE_EXISTING	Opens the file. Once opened, the file is truncated 
//										so that its size is zero bytes. The calling process must 
//										open the file with at least GENERIC_WRITE access. 
//										The function fails if the file does not exist.
//
//			shareMode - indicates how the file should be shared; it affects other applications
//						that will to access that file. Possible flags here are:
//							FILE_SHARE_DELETE Windows NT only: 
//										Subsequent open operations on the object will succeed only if delete access is requested. 
//							FILE_SHARE_READ Subsequent open operations on the object will succeed only 
//										if read access is requested. 
//							FILE_SHARE_WRITE Subsequent open operations on the object will succeed only 
//										if write access is requested. 
//
//
template < class AType >
AType * TMemoryMappedFileAllocator< AType >::Allocate( LPSTR fileName, DWORD howToCreate, DWORD shareMode, DWORD initialSize )
{
 	// at first create the desired file
	HANDLE hFile = CreateFile( fileName, /*GENERIC_WRITE | */GENERIC_READ,
						shareMode, NULL, howToCreate, 
						FILE_ATTRIBUTE_NORMAL, NULL );
	if( hFile == INVALID_HANDLE_VALUE )
		throw "CreateFile() failed.";//OtherTrouble( "CreateFile() failed." );

	// Get the file's size (lower 32-bits only)
	fAllocationSize = initialSize == 0 ? GetFileSize( hFile, NULL ) : initialSize;

	// create the file-mapping object,
	HANDLE hFileMap = CreateFileMapping( hFile, NULL, /*PAGE_READWRITE*/PAGE_READONLY, 0, fAllocationSize, NULL );
	if( hFileMap == NULL )
	{
		CloseHandle( hFile );
		throw "CreateFileMapping() failed.";//OtherTrouble( "CreateFileMapping() failed." );
	}

	// close the file handle as soon as possible
	CloseHandle( hFile );

	// now, map the file to the process' address space 
	AType * dataArray = (AType *)MapViewOfFile( hFileMap, /*FILE_MAP_WRITE*/FILE_MAP_READ, 0, 0, 0 );
	if(  dataArray == NULL )
		throw "MapViewOfFile() failed"; //OtherTrouble( "MapViewOfFile() failed" );

	// close the file-mapped object handle as soon as possible
	CloseHandle( hFileMap );

	return dataArray;

}

template < class AType >
void TMemoryMappedFileAllocator< AType >::Free( AType * mappedPtr )
{
	if ( mappedPtr != NULL )
		UnmapViewOfFile( mappedPtr );
}



///////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////
// Template declaration.

template <class AType > class SaveMemoryMappedFileAllocator : private TMemoryMappedFileAllocator< AType >
{
	public:

		SaveMemoryMappedFileAllocator();						// class default constructor
		SaveMemoryMappedFileAllocator( const string & );		// class constructor
		virtual ~SaveMemoryMappedFileAllocator();				// class virtual destructor

		AType *	Allocate( const string & fileName, DWORD howToCreate = OPEN_EXISTING, DWORD shareMode = FILE_SHARE_READ | FILE_SHARE_WRITE /*| FILE_SHARE_DELETEadded by BC on July 30th'98*/, DWORD initialSize = 0 );
		AType *	Allocate( LPSTR fileName, DWORD howToCreate = OPEN_EXISTING, DWORD shareMode = FILE_SHARE_READ | FILE_SHARE_WRITE /*| FILE_SHARE_DELETEadded by BC on July 30th'98*/, DWORD initialSize = 0 );
		
		AType *	GetAllocation( void ) const		{ return fAllocationPtr; }
		DWORD	GetAllocationSize( void ) const { return fAllocationSize; }

		const string &	GetFileName( void ) const		{ return fFileName; }

	protected:

		// Protected new operator to disable heap allocations of this object by a user.
		void *	operator new ( size_t ) { return ::new SaveMemoryMappedFileAllocator< AType >(); }

	private:

		AType *	fAllocationPtr;
		string	fFileName;
};


///////////////////////////////////////////////////////////////////////////////////////
// Template implementation.


template <class AType > 
SaveMemoryMappedFileAllocator< AType > :: SaveMemoryMappedFileAllocator()
{
	fAllocationPtr = NULL;
	fFileName = string("");
}

template <class AType > 
SaveMemoryMappedFileAllocator< AType > :: SaveMemoryMappedFileAllocator( const string & fileName )
{
	Allocate( fileName );		// delegate the action to one of its members
}

template <class AType > 
SaveMemoryMappedFileAllocator< AType > :: ~SaveMemoryMappedFileAllocator()
{
	Free( fAllocationPtr );
}


///////////////////////
template <class AType > 
AType * SaveMemoryMappedFileAllocator< AType > :: Allocate( const string & fileName, DWORD howToCreate, DWORD shareMode, DWORD initialSize )
{
	fFileName = fileName;
	return fAllocationPtr = TMemoryMappedFileAllocator< AType >::Allocate( (LPSTR)fileName.c_str(), howToCreate, shareMode, initialSize );
}

template <class AType > 
AType * SaveMemoryMappedFileAllocator< AType > :: Allocate(  LPSTR fileName, DWORD howToCreate, DWORD shareMode, DWORD initialSize )
{
	fFileName = string( fileName );
	return fAllocationPtr = TMemoryMappedFileAllocator< AType >::Allocate( fileName, howToCreate, shareMode, initialSize );
}


///////////////////////////////////////////////////////////////////////////////////////







#endif TPG_MEMORY_MAPPED_FILE_ALLOCATOR_H





